from prophet import Prophet

from prophet.diagnostics import cross_validation
from prophet.diagnostics import performance_metrics
from prophet.utilities import regressor_coefficients

from tabulate import tabulate
from sklearn.metrics import r2_score, mean_absolute_error

import pandas as pd
from pandas import DataFrame

import logging

from tscausalinference.evaluators import mape

logger = logging.getLogger('cmdstanpy')
logger.addHandler(logging.NullHandler())
logger.propagate = False
logger.setLevel(logging.CRITICAL)

def synth_analysis(df: DataFrame = None, 
                    regressors: list = [], 
                    intervention: list = None, 
                    seasonality: bool = True,
                    cross_validation_steps: int = 5
                    ):
    """
    A function to analyze a dataset using the Prophet library and provide evaluation metrics, response metrics and 
    intervention metrics. 

    Args:
    ------
    df : DataFrame, default None
        A pandas DataFrame with two columns 'ds' and 'y', containing the date and the corresponding target variable.
    regressors : list, default []
        List of column names in df to be used as regressors for the Prophet model.
    intervention : list, default None
        A list with two values, containing the start and end date for the intervention period, as a string in the 
        format 'YYYY-MM-DD'.
    seasonality : bool, default True
        Whether or not to include seasonality in the Prophet model.
    cross_validation_steps : int, default 5
        The number of days to use for cross-validation.
    
    Returns:
    -------
    data : DataFrame
    prints :
        - Cross-validation MAPE
        - Regressor coefficients (if any)
        - Pre-intervention response metrics (r2, MAE, and MAPE)
        - Intervention metrics (actual and predicted cumulative, and difference)

    Raises:
    -------
    ValueError: 
        If df is not a DataFrame, or does not contain columns named 'ds' and 'y'.
        If intervention is not a list with two values.
        If intervention period is not found in df.
    TypeError: 
        If df is None.
    KeyError:
        If any regressor in the list is not a column in df.

    Examples:
    --------
    >>> synth_analysis(df, ['regressor_1', 'regressor_2'], ['2021-01-01', '2021-02-01'], False, 10)
    Training period: 2019-01-01 to 2021-01-01
    Test period: 2021-01-01 to 2021-02-01

    Cross-validation MAPE: 5.83%
    +---------------+------------------+
    | Regressor     |   Coefficient    |
    +===============+==================+
    | regressor_1   |          0.293   |
    +---------------+------------------+
    | regressor_2   |         -0.0133  |
    +---------------+------------------+

    """

    if not isinstance(df, pd.DataFrame):
        raise ValueError("df must be a pandas DataFrame")
    elif not set(['ds', 'y']).issubset(df.columns):
        raise ValueError("df must contain columns named 'ds' and 'y'")
    elif df.empty:
        raise TypeError("df is empty")

    if isinstance(intervention, type(None)):
      intervention_error = 'Define your intervention as a list with the start_date and end_date of your intervention'
      raise intervention_error
    
    pre_intervention = [df.ds.min(),(pd.to_datetime(intervention[0]) - pd.Timedelta(days=1)).strftime('%Y-%m-%d')]
    post_intervention = [(pd.to_datetime(intervention[0]) + pd.Timedelta(days=1)).strftime('%Y-%m-%d'),df.ds.max()]

    training_dataframe = df[(df.ds > pd.to_datetime(pre_intervention[0]))&(df.ds <= pd.to_datetime(pre_intervention[1]))&(df.y > 0)].fillna(0).copy()
    training_dataframe['ds'] = pd.to_datetime(training_dataframe['ds'])

    test_dataset = df[(df.ds > pd.to_datetime(pre_intervention[0]))&(df.ds <= pd.to_datetime(post_intervention[1]))&(df.y > 0)].fillna(0).copy()
    test_dataset['ds'] = pd.to_datetime(test_dataset['ds'])

    prediction_period = len(test_dataset[(test_dataset.ds > pd.to_datetime(intervention[0]))&(test_dataset.ds <= pd.to_datetime(intervention[1]))].index)

    print('Training period: {} to {}'.format(pre_intervention[0], pre_intervention[1]))
    print('Test period: {} to {}\n'.format(intervention[0], intervention[1]))

    if seasonality:
        prophet = Prophet(daily_seasonality=True, 
                    weekly_seasonality=True, 
                    changepoint_prior_scale = 0.05, 
                    changepoint_range = 0.85)
    else:
        prophet = Prophet(changepoint_prior_scale = 0.05, 
                    changepoint_range = 0.85)

    for regressor in regressors:
            prophet.add_regressor(name = regressor)
    
    prophet.fit(training_dataframe)

    prophet_predict = prophet.predict(test_dataset)

    df_cv = cross_validation(prophet, horizon = '{} days'.format(cross_validation_steps),disable_tqdm=False)
    df_p = performance_metrics(df_cv)

    model_mape_mean = df_p.mape.mean()

    df['ds'] = pd.to_datetime(df['ds'])

    data = pd.merge(
        prophet_predict[['ds','yhat', 'yhat_lower', 'yhat_upper']], 
        df[["ds", "y"]], how='left', on='ds'
        )

    data['yhat'] = data['yhat'].astype(float)
    data['y'] = data['y'].astype(float)

    data['cummulitive_y'] = data['y'].cumsum()
    data['cummulitive_yhat'] = data['yhat'].cumsum()

    data['point_effects'] = data['y'] - data['yhat']
    data['cummulitive_effect'] = data['point_effects'].cumsum()

    data['cummulitive_yhat_lower'] = data['yhat_lower'].cumsum()
    data['cummulitive_yhat_upper'] = data['yhat_upper'].cumsum()

    # print response
    print(f'Cross-validation MAPE: {model_mape_mean:.2%}')
    if len(regressors) >= 1:
        regressor_df = regressor_coefficients(prophet)
        # format table
        table = []
        for index, row in regressor_df.iterrows():
            table.append([index, row])

        print(tabulate(table, headers=['Regressor', 'Coefficient'], tablefmt='grid'))

    pre_int_metrics = [
    ['r2', r2_score(y_pred = data[(data.ds > pd.to_datetime(pre_intervention[0]))&(data.ds <= pd.to_datetime(pre_intervention[1]))&(data.y > 0)].yhat, y_true = data[(data.ds > pd.to_datetime(pre_intervention[0]))&(data.ds <= pd.to_datetime(pre_intervention[1]))&(data.y > 0)].y)],
    ['MAE', mean_absolute_error(y_pred = data[(data.ds > pd.to_datetime(pre_intervention[0]))&(data.ds <= pd.to_datetime(pre_intervention[1]))&(data.y > 0)].yhat, y_true = data[(data.ds > pd.to_datetime(pre_intervention[0]))&(data.ds <= pd.to_datetime(pre_intervention[1]))&(data.y > 0)].y)],
    ['MAPE (%)', mape(y_pred = data[(data.ds > pd.to_datetime(pre_intervention[0]))&(data.ds <= pd.to_datetime(pre_intervention[1]))&(data.y > 0)].yhat, y_true = data[(data.ds > pd.to_datetime(pre_intervention[0]))&(data.ds <= pd.to_datetime(pre_intervention[1]))&(data.y > 0)].y)]
    ]

    strings_info = """
+----------------------+
Pre intervention metrics
+----------------------+
{}
    """

    print(
        strings_info.format(
            tabulate(pre_int_metrics, 
            headers=['Metric', 'Value'], 
            tablefmt='pipe')
        ).strip()
    )

    int_metrics = [
    ['Actual cumulative', data[(data.ds >= intervention[0]) & (data.ds < intervention[1])].y.sum()],
    ['Predicted cumulative:', data[(data.ds >= intervention[0]) & (data.ds < intervention[1])].yhat.sum()],
    ['Difference', data[(data.ds >= intervention[0]) & (data.ds < intervention[1])].point_effects.sum()],
    ['Change (%)', (data[(data.ds >= intervention[0]) & (data.ds < intervention[1])].y.sum() / data[(data.ds >= intervention[0]) & (data.ds < intervention[1])].yhat.sum() -1)*100]
    ]
    
    return data, pre_int_metrics, int_metrics


import pandas as pd
import numpy as np

from typing import Union

def bootstrap_simulate(
                    data: Union[np.array, pd.DataFrame] = None, 
                    n_samples: int = 1500, 
                    n_steps: int = None):
    """
    Simulates bootstrap samples using a random walk process based on the input data.
    
    The function follows the principles of bootstrapping, which is a resampling method used to estimate the distribution of a statistic from a sample of data. Bootstrapping involves repeatedly resampling the original data with replacement to create new samples of the same size as the original sample. Each new sample is used to calculate the statistic of interest, and the distribution of these calculated statistics is used to estimate the distribution of the statistic in the population.

    Args:
        data: A 1D array-like object containing the original data.
        n_samples: An integer representing the number of bootstrap samples to generate.
        n_steps: An integer representing the number of steps in the random walk process.

    Returns:
        An array of shape (n_samples, n_steps) containing the bootstrap samples generated by the random walk process.
    """
    # Initialize array to hold bootstrap samples
    bootstrap_samples = np.empty((n_samples, n_steps))
    
    # Loop over number of bootstrap samples
    for i in range(n_samples):
        # Resample data with replacement
        bootstrap_data = np.random.choice(data, size=len(data))#, replace=True)
        
        # Simulate random walk based on bootstrap data
        walk = np.cumsum(np.random.randn(n_steps))
        walk -= walk[0]
        walk *= bootstrap_data.std() / walk.std()
        walk += bootstrap_data.mean()
        
        # Save random walk as one of the bootstrap samples
        bootstrap_samples[i] = walk
    
    return bootstrap_samples
    
def bootstrap_p_value(
                    control: Union[np.array, pd.DataFrame] = None, 
                    treatment: Union[np.array, pd.DataFrame] = None, 
                    simulations: np.array = None, 
                    alpha: float = 0.05):
    """
    Calculate the p-value and false positive rate for a two-sample hypothesis test using bootstrapping
    with a specified number of simulations.

    Args:
        control: A 1-D array-like object of control group data.
        treatment: A 1-D array-like object of treatment group data.
        simulations: A 2-D array-like object containing bootstrap simulations of the data.
        center: A boolean indicating whether to center the data around the control mean (default False).
        alpha: The significance level for the hypothesis test (default 0.05).

    Returns:
        A tuple containing the p-value and false positive rate, the lower and upper bounds of the confidence
        interval, and the bootstrapped means of the simulations.

    Raises:
        ValueError: If either control or treatment is empty, or if simulations is empty.

    Examples:
        >>> control = [1, 2, 3, 4, 5]
        >>> treatment = [3, 4, 5, 6, 7]
        >>> bs_sims = [np.random.choice(control, size=len(control), replace=True) for i in range(1000)]
        >>> simulations_bootstrap_p_value(control, treatment, bs_sims, center=True, alpha=0.01)
        ([0.018, 0.0], [-3.125, 2.875], array([-1.8, -1.2, -1.2, ...,  1.8,  1.8,  1.8]))
    """
        
    # Calculate the mean of the data
    mean = np.mean(control)
    mean_treatment = np.mean(treatment)
        
    bootstrapped_means = np.empty(len(simulations))
    
    for i in range(len(simulations)):
        bootstrapped_means[i] = simulations[i].mean()
    
    lower, upper = np.percentile(bootstrapped_means, [alpha / 2 * 100, (1 - alpha / 2) * 100])
    
    fp_rate = len(bootstrapped_means[(bootstrapped_means >= (- mean_treatment)) & 
                                            (bootstrapped_means <= mean_treatment)]) / len(simulations)
    
    p_mean_negative = len(bootstrapped_means[bootstrapped_means < (- mean_treatment)]) / len(simulations)
    p_mean_positive = len(bootstrapped_means[bootstrapped_means > mean_treatment]) / len(simulations)
    
    p_value = p_mean_positive + p_mean_negative
    return [p_value, fp_rate], [lower, upper], bootstrapped_means


from typing import List, Union

import sys, os, io
from pandas import DataFrame
import pandas as pd

import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from tabulate import tabulate

from statsmodels.tsa.seasonal import seasonal_decompose

from tscausalinference.synth_regression import synth_analysis
from tscausalinference.bootstrap import bootstrap_simulate, bootstrap_p_value

sns.set_theme()
sns.set_context("paper")

class tscausalinference:
    """
   Time series causal inference using structural causal models (SCM) and a difference-in-differences (DiD) approach.

    Parameters
    ----------
    data : Union[np.array, DataFrame]
        A NumPy array or Pandas DataFrame containing the time series data.
    intervention : Union[List[int], List[str], List[pd.Timestamp]]
        A list of two elements indicating the start and end dates of the intervention.
    regressors : list, optional
        A list of column names representing regressors in the SCM. Default is an empty list.
    alpha : float, optional
        The level of significance for the hypothesis test. Default is 0.05.
    past_window : int, optional
        The number of days to show in the plot before the start of the intervention. Default is 5.
    back_window : int, optional
        The number of days to show in the plot after the end of the intervention. Default is 25.
    seasonality : bool, optional
        If True, the model includes a seasonal component. Default is True.
    n_samples : int, optional
        The number of bootstrap samples to simulate. Default is 1500.
    cross_validation_steps : int, optional
        The number of cross-validation steps used for training. Default is 5.

    Methods
    -------
    plot_intervention()
        Plots the time series data before and after the intervention with confidence intervals.
    plot_simulations(simulation_number: int = 10)
        Plots a specified number of bootstrap simulations.
    """

    def __init__(self,
        data: Union[np.array, DataFrame],
        intervention: Union[List[int], List[str], List[pd.Timestamp]],
        regressors: list = [],
        alpha: float = 0.05,
        past_window: int = 5,
        back_window: int = 25,
        seasonality: bool = True,
        n_samples: int = 1500,
        cross_validation_steps: int = 5
        ):
        """
        """

        self.data = data
        self.intervention = intervention
        self.alpha = alpha
        self.past_window = past_window
        self.back_window = back_window
        self.seasonality = seasonality
        self.regressors = regressors
        self.n_samples = n_samples
        self.cross_validation_steps = cross_validation_steps

        self.data, self.pre_int_metrics, self.int_metrics = synth_analysis(
            df = data, 
            regressors = regressors, 
            intervention = intervention, 
            seasonality = seasonality,
            cross_validation_steps = cross_validation_steps
            )
        self.string_filter = "ds >= '{}' & ds <= '{}'".format(intervention[0],intervention[1])
        
        self.simulations = bootstrap_simulate(
                data = self.data.query(self.string_filter).yhat, 
                n_samples = n_samples, 
                n_steps = len(self.data.query(self.string_filter).index)
                )
        
        self.stadisticts, self.stats_ranges, self.samples_means = bootstrap_p_value(control = self.data.query(self.string_filter).yhat, treatment = self.data.query(self.string_filter).y, simulations = self.simulations)
 
    def plot_intervention(self):
        """
        The function plot_intervention is a method of the tscausalinference class. It generates a plot showing the predicted and actual values of the target variable around the intervention period, as well as the cumulative effect of the intervention.

        Parameters
        ----------
            No parameters are required.
        
        Raises
        -------
            No raises are defined.
        
        Returns
        --------
            No returns are defined, as the method simply generates a plot.
        """
        data = self.data
        fig, axes = plt.subplots(nrows=2, ncols=1, figsize=(15, 10))

        lineplot = sns.lineplot(x = 'ds', y = 'yhat', color = 'r', alpha=0.5, linestyle='--', ci=95,
                            err_kws={'linestyle': '--', 'hatch': '///', 'fc': 'none'}, ax=axes[0],
                    data = data[(data.ds >= pd.to_datetime(self.intervention[0]) - pd.Timedelta(days=self.back_window))&(data.ds <= pd.to_datetime(self.intervention[1]) + pd.Timedelta(days=self.past_window))]
                    )

        sns.lineplot(x = 'ds', y = 'y', ax=axes[0], color = 'b',
                    data = data[(data.ds >= pd.to_datetime(self.intervention[0]) - pd.Timedelta(days=self.back_window))&(data.ds <= pd.to_datetime(self.intervention[1]) + pd.Timedelta(days=self.past_window))]
                    )

        lineplot.axvline(pd.to_datetime(self.intervention[0]), color='r', linestyle='--',alpha=.5)
        lineplot.axvline(pd.to_datetime(self.intervention[1]), color='r', linestyle='--',alpha=.5)

        lineplot.fill_between(data[(data.ds >= pd.to_datetime(self.intervention[0]) - pd.Timedelta(days=self.back_window))&(data.ds <= pd.to_datetime(self.intervention[1]) + pd.Timedelta(days=self.past_window))]['ds'], 
                        data[(data.ds >= pd.to_datetime(self.intervention[0]) - pd.Timedelta(days=self.back_window))&(data.ds <= pd.to_datetime(self.intervention[1]) + pd.Timedelta(days=self.past_window))]['yhat_lower'],
                        data[(data.ds >= pd.to_datetime(self.intervention[0]) - pd.Timedelta(days=self.back_window))&(data.ds <= pd.to_datetime(self.intervention[1]) + pd.Timedelta(days=self.past_window))]['yhat_upper'], 
                        color='r', alpha=.1)

        lineplot.legend(['Prediction', 'Real'])

        cumplot = sns.lineplot(x = 'ds', y = 'cummulitive_effect', color = 'g',
             data = data[(data.ds >= pd.to_datetime(self.intervention[0]) - pd.Timedelta(days=self.back_window))&(data.ds <= pd.to_datetime(self.intervention[1]) + pd.Timedelta(days=self.past_window))],
             ax=axes[1]
             )

        sns.lineplot(x = 'ds', y = 'cummulitive_effect', color = 'b', linestyle='--',
                    data = data[(data.ds >= self.intervention[0]) & (data.ds <= self.intervention[1])],
                    ax=axes[1]
                    )

        sns.lineplot(x = 'ds', y = 'cummulitive_effect', color = 'g',
                    data = data[data.ds > self.intervention[1]],
                    ax=axes[1]
                    )

        cumplot.axvline(pd.to_datetime(self.intervention[0]), color='r', linestyle='--')
        cumplot.axvline(pd.to_datetime(self.intervention[1]), color='r', linestyle='--')

        cumplot.axvspan(pd.to_datetime(self.intervention[0]), pd.to_datetime(self.intervention[1]), alpha=0.07, color='r')

        plt.show()

    def plot_simulations(self, simulation_number: int = 10):
        """
        The plot_simulations() method of tscausalinference class plots the distribution of the mean difference between the treatment and control group based on the bootstrap simulations generated in bootstrap_simulate(). 
        The plot includes a histogram and a box plot of the simulated means.

        Parameters
        ----------
            Simulation_number (int): The number of simulations to plot. Default is 10.
        
        Raises
        -------
            ValueError: if simulation_number is greater than the number of simulations generated.
        """
        data = self.data

        fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(18, 5))
        for i in range(simulation_number):#range(len(samples[0])):
            sns.lineplot(x=data[(data.ds >= pd.to_datetime(self.intervention[0]))&(data.ds <= pd.to_datetime(self.intervention[1]))].ds, y=self.simulations[i], 
                        linewidth=0.5, alpha=0.45,
                        color = 'orange', legend = False, ax=axes[0])

        sns.lineplot(x = 'ds', y = 'yhat', color = 'b',
                    err_kws={'linestyle': '--', 'hatch': '///', 'fc': 'none'}, ax=axes[0],
                    data = data[(data.ds <= pd.to_datetime(self.intervention[0]))],
                    linewidth=1, label='Training')

        sns.lineplot(x = 'ds', y = 'yhat', color = 'b',
                    err_kws={'linestyle': '--', 'hatch': '///', 'fc': 'none'},
                    data = data[(data.ds >= pd.to_datetime(self.intervention[0]))&(data.ds <= pd.to_datetime(self.intervention[1]))],
                    linewidth=1, label='Forcast', ls='--', ax=axes[0])

        sns.lineplot(x = 'ds', y = 'y', color = 'g',
                    err_kws={'linestyle': '--', 'hatch': '///', 'fc': 'none'},
                    data = data[(data.ds >= pd.to_datetime(self.intervention[0]))&(data.ds <= pd.to_datetime(self.intervention[1]))],
                    label='Real', ax=axes[0])
        
        # add a custom legend entry for the yellow line
        custom_legend = plt.Line2D([], [], color='orange', label='simulations')

        # create a new legend with your custom entry and add it to the plot
        handles, labels = axes[0].get_legend_handles_labels()
        handles.append(custom_legend)
        axes[0].legend(handles=handles, loc='upper left')

        # Add the mean value to the right corner
        plt.text(1.05, 0.95, f'P-Value: {self.stadisticts[0]:.2f}', ha='left', va='center', transform=plt.gca().transAxes)
        plt.text(1.05, 0.80, f'P-Effect: {self.stadisticts[1]:.2f}', ha='left', va='center', transform=plt.gca().transAxes)

        sns.histplot(self.samples_means, kde=True, ax=axes[1])

        # Add title and labels
        plt.title('Histogram of Bootstrapped Means')
        plt.xlabel('Bootstrapped Mean Value')
        plt.ylabel('Frequency')

        plt.axvline(self.stats_ranges[0], color='r', linestyle='--',alpha=.5,)
        plt.axvline(self.stats_ranges[1], color='r', linestyle='--',alpha=.5,)

        # Add the real mean as a scatter point
        plt.scatter(data[(data.ds >= pd.to_datetime(self.intervention[0]))&(data.ds <= pd.to_datetime(self.intervention[1]))].y.mean(), 0, color='orange', s=600)

        # Show the plot
        sns.despine()
    
    def summary(self):
        """
        Parameters
        ----------
            No parameters are required.
        
        Raises
        -------
            No raises are defined.
        
        Returns
        --------
            No returns are defined, as the method simply generates a overview.
        """
        data = self.data
        
        data['ds'] = pd.to_datetime(data['ds'])
        x_decomp = data[(data.ds <= pd.to_datetime(self.intervention[0]))][['ds', 'y']].set_index('ds')

        decomposition_obj = seasonal_decompose(
                    x = x_decomp, 
                    model = 'additive'
                    )
        
        std_res = decomposition_obj.resid.describe()['std']

        if std_res < 0.2:
            mde = 15
            noise = 'low'
        elif (std_res > 0.2) & (std_res <= 0.6):
            mde = 21
            noise = 'medium'
        elif std_res > 0.6:
            mde = 25
            noise = 'high'
        
        if self.stadisticts[0] <= 0.01:
            summary = """
    Considerations
    --------------
    a) The standard deviation of the residuals is {}. This means that the noise in your data is {}.
    b) Based on this information, in order for your effect to be detectable, it should be greater than {}%.

    Summary
    -------
    During the intervention period, the response variable had an average value of approximately {}. 
    By contrast, in the absence of an intervention, we would have expected an average response of {}. 
    The 95% confidence interval of this counterfactual prediction is {} to {}.

    The usual error of your model is {}%, while the difference during the intervention period is {}%. 
    During the intervention, the error was {}%, suggesting some factor is impacting the quality of the model,
    and that the differences significant.

    The probability of obtaining this effect by chance is very small 
    (after {} simulations, bootstrap probability p = {}). 
    This means that the causal effect can be considered statistically significant.
            """
        else:
            summary = """
    Considerations
    --------------
    a) The standard deviation of the residuals is {}. This means that the noise in your data is {}.
    b) Based on this information, in order for your effect to be detectable, it should be greater than {}%.

    Summary
    -------
    During the intervention period, the response variable had an average value of approximately {}. 
    By contrast, in the absence of an intervention, we would have expected an average response of {}. 
    The 95% confidence interval of this counterfactual prediction is {} to {}.

    The usual error of your model is {}%, while the difference during the intervention period is {}%. 
    During the intervention, the error was {}%, suggesting that the model can explain well what should happen,
    and that the differences are not significant.

    The probability of obtaining this effect by chance is not small 
    (after {} simulations, bootstrap probability p = {}). 
    This means that the causal effect cannot be considered statistically significant.
            """

        print(
            summary.format(
                round(std_res,2),
                noise,
                mde,
                round(data[(data.ds >= pd.to_datetime(self.intervention[0])) & (data.ds <= pd.to_datetime(self.intervention[1]))].y.mean(),2),
                round(data[(data.ds >= pd.to_datetime(self.intervention[0])) & (data.ds <= pd.to_datetime(self.intervention[1]))].yhat.mean(),2),
                round(data[(data.ds >= self.intervention[0]) & (data.ds <= self.intervention[1])].yhat_lower.mean(),2), 
                round(data[(data.ds >= self.intervention[0]) & (data.ds <= self.intervention[1])].yhat_upper.mean(),2),
                round(self.pre_int_metrics[2][1],2),
                round(self.int_metrics[3][1],2),
                round(self.int_metrics[3][1]/self.pre_int_metrics[2][1],2),
                self.n_samples,
                self.stadisticts[0]
            )
        )
    
    def summary_intervention(self):
        """
        Parameters
        ----------
            No parameters are required.
        
        Raises
        -------
            No raises are defined.
        
        Returns
        --------
            No returns are defined, as the method simply generates a overview.
        """
        data = self.data

        strings_info = """
+-----------------------+-----------+
      Pre intervention metrics
+-----------------------+-----------+
{}
+-----------------------+-----------+
      {}
        """

        print(
            strings_info.format(
                tabulate(self.int_metrics, 
                headers=['Metric', 'Value'], 
                tablefmt='pipe'),
                'CI 95%: [{}, {}]'.format(
                    round(data[(data.ds >= self.intervention[0]) & (data.ds <= self.intervention[1])].yhat_lower.sum(),2), 
                    round(data[(data.ds >= self.intervention[0]) & (data.ds <= self.intervention[1])].yhat_upper.sum()),2)
            ).strip()
        )
